#include "atari.h"
#include "config.h"
#include "monitor.h"
#include "pokeysnd.h"
#include "sound.h"

#include <stdlib.h>
#include <string.h>

#include "atari_gp32.h"

// We have dual sound buffers
// i) Controlled by gpsoundbuffer.c. Callback when we need more data to play.
// ii) Controlled by emulator. Filled every frame.
//
// Strategy: Fill emulator buffer in main code. Copy into play buffer on callback.
//
// Reason: Emulator does not like callbacks on interrupt into the sound code.
// Mostly works, but volume only seems broken among other things.
// Quite interesting really since the SDL port appears to use the callback method only.
// Another reason: Decouples emulator time from GP32 time

int soundBufferSamplesStored = 0;

#define DEFDSPRATE 22050
static GPSOUNDBUF soundBuffer;
static int soundPlaying = FALSE;

#ifdef STEREO
	#define SNDBUFSIZE 10000
#else
	#define SNDBUFSIZE 5000
#endif
static char * atariSoundBuffer = 0;
static int atariSoundBufferStart = 0;
static int atariSoundBufferEnd = 0;

//////////////////////////////////////////////////////////////////////////
// Forward declarations of some internal functions
static void gpBufferCallback(void * userdata, u8 * stream, int len);
static int soundBufferRollingDistance(int start, int end, int length);

//////////////////////////////////////////////////////////////////////////
// Functions called externally by the emulator code...
void Sound_Initialise(int *argc, char *argv[])
{
	// And sound
#ifdef STEREO
	stereo_enabled = 0;
#endif

	atariSoundBuffer = (char *)malloc(SNDBUFSIZE);
#ifdef STEREO
        GpPcmInit(PCM_S22, PCM_8BIT);
	Pokey_sound_init(FREQ_17_EXACT, DEFDSPRATE, 2);
	soundBuffer.freq = PCM_S22;
#else
        GpPcmInit(PCM_M22, PCM_8BIT);
	Pokey_sound_init(FREQ_17_EXACT, DEFDSPRATE, 1);
	soundBuffer.freq = PCM_M22;
#endif
	soundBuffer.format = PCM_8BIT;
	soundBuffer.pollfreq = 100;
	soundBuffer.timerId = gpNextTimer++;
	soundBuffer.callback = gpBufferCallback;
	soundBuffer.samples = 250;
}

void Sound_Pause(void)
{
	if (!soundPlaying)
		return;

	soundPlaying = FALSE;

	GpSoundBufStop();
}

void Sound_Continue(void)
{
	if (soundPlaying)
		return;

	soundPlaying = TRUE;	

	GpSoundBufStart(&soundBuffer);
}

void Sound_Exit(void)
{
	Sound_Pause();
}

void Sound_Update()
{
	if (soundPlaying)
	{
		// Fill buffer
		int samplesToGet = 0;
		int samplesLeft = 0;
		samplesLeft = soundBufferRollingDistance(atariSoundBufferStart, atariSoundBufferEnd, SNDBUFSIZE);

#ifdef STEREO
		samplesToGet = 22050*deltatime*2; // 22050/50 *2 approx
#else
		samplesToGet = 22050*deltatime;
#endif

		if (samplesLeft < samplesToGet)
		{
			samplesToGet *= 1.1;
		}
		else if (samplesLeft > (SNDBUFSIZE-samplesToGet))
		{
			samplesToGet *= 0.9;
		}
		if (samplesToGet&1)
			samplesToGet-=1;

		if (samplesToGet > 0)
		{
			if (SNDBUFSIZE < (atariSoundBufferEnd + samplesToGet)) //overflow
			{
				int remaining = SNDBUFSIZE - atariSoundBufferEnd;
				if (remaining>0)
					Pokey_process(atariSoundBuffer + atariSoundBufferEnd, remaining);
				if ((samplesToGet-remaining)>64)
				{
					Pokey_process(atariSoundBuffer, samplesToGet-remaining);
					atariSoundBufferEnd = samplesToGet-remaining;
				}
				else
					atariSoundBufferEnd = 0;
			}
			else
			{
				Pokey_process(atariSoundBuffer + atariSoundBufferEnd, samplesToGet);
				atariSoundBufferEnd += samplesToGet;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////
// Helper function for sound circular buffer
static int soundBufferRollingDistance(int start, int end, int length)
{
	if (end > start)
		return (end-start);
	if (end < start)
		return (length-start + end);

	return 0;
}

// Interrupt driven callback from the sound buffer code
static void gpBufferCallback(void * userdata, u8 * stream, int len)
{
	static int last = 0;
	
	int samplesAvailable = soundBufferRollingDistance(atariSoundBufferStart,atariSoundBufferEnd, SNDBUFSIZE);
	if (samplesAvailable>len)
	{
		if ((atariSoundBufferStart+len) < SNDBUFSIZE)
		{
			memcpy(stream, atariSoundBuffer+atariSoundBufferStart, len);
			atariSoundBufferStart+=len;
		}
		else
		{
			int remaining = SNDBUFSIZE - atariSoundBufferStart;
			memcpy(stream, atariSoundBuffer+atariSoundBufferStart, remaining);

			int fromStart = len-remaining;
			if (fromStart)
			{
				memcpy(stream+remaining, atariSoundBuffer, fromStart);
			}
			atariSoundBufferStart = fromStart;
		}
	}
	else
	{
		memset(stream, last, len);
	}

	last = stream[len-1];

	// Store count of samples -> for external use in display status text
	soundBufferSamplesStored = samplesAvailable;
	
//	debugText = debugTextBuffer;
//	sprintf(debugTextBuffer,"%4d           ", soundBufferSamplesStored);
	
//	if (soundBufferSamplesStored < 500)
//		sprintf(debugTextBuffer+6, "ALERT");
}

